function outmesh=meshP1P2(node,edge,edgeCL,hdata,options)
%--------------------------------------------------------------------------
% MAILLAGES P1 et P2 (cf. doc_maillageP1P2.pdf)
%--------------------------------------------------------------------------
%
% * Maillage P1 :
%
%   v1 : coordonnees des noeuds v1=[x, y, label]
%        label : reference des noeuds :
%                label=0 -> noeud du domaine interieur
%                label=i -> noeud sur le bord defini par edge
%
%   e1 : aretes e1=[n1, n2, label, t1, t2] 
%        - (n1,n2) sont les deux noeuds de l'arete
%        - (t1,t2) sont les deux triangles de part et d'autre de l'arete 
%          (t2=0, si l'arete est sur le bord).
%
%   t1 : triangles t1=[n1,n2,n3, label,aire, e1,e2,e3]
%        - (n1,n2,n3) trois sommets des triangles
%        - (e1,e2,e3) trois aretes  des triangles
%
%
% * Maillage P2 :
%
%   v2,e2 : coordonnees des noeuds et aretes (meme structure que v1,e1)
%
%   t2    : triangles obtenus en coupant en 4 les triangles du maillage P1.
%           t2 = [n1, m2, m3, label,aire, e1,e2,e3;
%                 n2, m1, m3, label,aire, e1,e2,e3;
%                 n3, m2, m1, label,aire, e1,e2,e3;
%                 m1, m2, m3, label,aire, e1,e2,e3]
%
%           ou m1,m2,m3 sont les points milieux des aretes des triangles
%           P1; m1,m2,m3 sont sur les aretes opposees aux noeud n1,n2,n3
%           respectivement.
%
% * Supports (P1 et P2)
%
% isupp,tsupp : permet de donner la liste des triangles ayant un noeud 
%                 donne comme sommet (support).
%      isupp(i) : adresse dans le vecteur tsupp du premier triangle ayant 
%                 le noeud i comme sommet.
%                 La liste complete des triangles ayant le noeud i comme 
%                 sommet, est donnee par :
%
%                 tsupp(k) pour k = isupp(i) + 1, ... , isupp(i + 1).
%
%--------------------------------------------------------------------------

fprintf('Maillage mesh2d (v3.1) Copyright@2017 Darren Engwirda\n');
%fprintf('-----------------------------------------------------------\n');
options.kind = 'delfront';
options.rho2 = +1.0 ;
part=[];
[p,e,t,tnum] = refine2(node,edge,part,options,hdata.hmax);
opts.vtol= 1.0e-2;
opts.iter = +64; 
opts.disp = +16;
[p,e,t,tnum] = smooth2(p,e,t,tnum,opts);

% nv1 : nb de noeuds du maillage P1
nv1=size(p,1); 

% nt1 : nb de triangles du maillage P1
nt1=size(t,1); 

%--------------------------------------------------------------------------
% Construction etendue du maillage P1
%--------------------------------------------------------------------------
% Reference des noeuds :
%  0 -> noeud du domaine interieur
%  i -> noeud sur le bord defini par edge
%--------------------------------------------------------------------------

% References des noeuds interieures aux sous-domaines
labelnd=zeros(nv1,1);

% References des noeuds sur les bords (edge) avec Dirichlet
% Prise en compte des CL : si i noeud avec CL Neumann => label(i)=0
pa = node(edge(:,1),:);
pb = node(edge(:,2),:);
[lp,lj] = findline(pa,pb,p);
indbd = find(lp(:,1)~=0);
label = lj(lp(indbd,:));
for k=1:size(label,1)
    ind = find(ismember(label(k,:),edge(edgeCL)));
    if isempty(ind)
        labelnd(indbd(k))=label(k,1);
    else
        labelnd(indbd(k))=label(k,ind(1));
    end
end
    
v1=[p, labelnd];

% Construction d'autres donnees du maillage:

%    Aretes du maillage P1
t=[t, tnum];
e1=make_edges(v1,t);

%    Calcul des aires des triangles
t1=area_triangle(v1,t);

%    Recuperation des (trois) aretes de chaque triangle.
t1=edges_in_triangle(e1,t1);

%    Construction du support noeud -> triangle 
[isupp1,tsupp1]=make_supp(t1);

%--------------------------------------------------------------------------
% Construction du maillage P2 (a partir du maillage P1)
%--------------------------------------------------------------------------
[v2,t2]=P1toP2(v1,t1,e1);
e2=make_edges(v2,t2);
t2=area_triangle(v2,t2);
t2=edges_in_triangle(e2,t2);
[isupp2,tsupp2]=make_supp(t2);

% nv2 : nb de noeuds du maillage P2
nv2=size(v2,1); 

% Sommets de t2 associes aux triangles t1
t12 = [t1(:,1:3), t2(3*nt1+1:4*nt1,1:3)];

%--------------------------------------------------------------------------
% Indices des noeuds du bord
%--------------------------------------------------------------------------
ib2= find(ismember(v2(:,3),edge(edgeCL)));   %noeuds du bord
ic2=(1:nv2)';             
ic2(ib2)=[];           %suppression dans ic2 de tous les points du bord 
                       %ic2 : noeuds interieurs

ibT=[ib2; ib2+nv2];    %noeuds du bord pour les 2 composantes de la vitesse

outmesh=struct('v1',v1,'t1',t1,'e1',e1,'isupp1',isupp1,'tsupp1',tsupp1,...
    'v2',v2,'t2',t2,'e2',e2,'isupp2',isupp2,'tsupp2',tsupp2,'t12',t12,...
    'ib2',ib2,'ic2',ic2,'ibT',ibT);

end
