%-------------------------------------------------------------------------
% Solveur de Stokes - Elements finis P2/P1
%
%    -nu Delta u + grad(p) = f dans Omega polygonal
%     div(u)= 0
%
% Conditions limites :
%   Dirichlet : u=g sur une partie S1 du bord de Omega
%     Neumann : sigma(u,p)n = 0 sur une partie S2 (S1 U S2 = bord de Omega)
%
% Maillage par mesh2d v3.1.0 - GNU GPL - Copyright (C) 2017 Darren Engwirda
%
%-------------------------------------------------------------------------
close all;
% 
% Path pour mesh2d (maillage)
addpath ./mesh2d_v31
addpath ./lib

fprintf('-------------------------------------------------------------\n');
fprintf('                  Solveur Stokes (P2/P1)\n');
fprintf('-------------------------------------------------------------\n');

nu=1;           % viscosite
%--------------------------------------------------------------------------
% Definition de la geometrie
%--------------------------------------------------------------------------
% Bord exterieur du domaine

% % Exemple 1 : rectangle [0,L]x[0,H]
% L=2; H=1;
% node = [0 0; L 0; L H; 0 H];
% edge = [1 2; 2 3; 3 4; 4 1];
% edgeCL = [1,3,4];   % aretes pour les CL de Dirichlet

% Exemple 2 : domaine 'L-shaped'
L=2; H=1;
d1=1; d2=0.5;
node = [0 0; d1 0; d1 -d2; L -d2; L 0; L H; 0 H];  
edge = [1 2; 2 3; 3 4; 4 5; 5 6; 6 7; 7 1];
edgeCL = [1,2,4,5,6,7];   % aretes pour les CL de Dirichlet

% % Exemple 3 : rectangle [0,L]x[0,H] avec un obstacle disque/ellipse
% L=1; H=0.5;
% node = [0 0; L 0; L H; 0 H];
% edge = [1 2; 2 3; 3 4; 4 1];
% edgeCL = [1,3,4];   % aretes pour les CL de Dirichlet
% 
% % construction de l'obstacle (ellipse)
% pxy=[0.3; 0.2];     % position (x,y) du centre de l'ellipse
% lgxy=[0.05; 0.05];  % longueurs des demi-axes et orientation
% angle=0;            % orientation de l'ellipse
% 
% pas_hlocal=pas_h/4;
% np=-1; % nombre de points minimal sur l'ellipse (si >0)
% N= max(floor(2*pi*sqrt((lgxy(1).^2+lgxy(2)^2)/2)/pas_hlocal),np);
% theta = linspace(0,2*pi,N+1); 
% theta=theta(1:N);
% x = lgxy(1)*cos(theta);
% y = lgxy(2)*sin(theta);
% Rotation=[cos(angle) -sin(angle); sin(angle) cos(angle)]*[x; y];
% x=pxy(1)+Rotation(1,:); 
% y=pxy(2)+Rotation(2,:);
%      
% % Construction de la structure pour le mailleur mesh2d
% n=size(node,1);
% node = [node; x' y'];
% edge = [edge; (n+1:n+N-1)', (n+2:n+N)'; n+N, n+1];
% edgeCL = [edgeCL, (n+1:n+N)];

%--------------------------------------------------------------------------
% Taille (pas) du maillage 
pas_h = 0.05;   
hdata = [];
hdata.hmax  = pas_h; 
optionmesh=[];
optionmesh.output=true;

%--------------------------------------------------------------------------
% Choix du solveur (systme linaire)
% sdr=1 : solveur direct; sdr=0 : solveur iteratif MINRES (cf README.txt)
%--------------------------------------------------------------------------
sdr=1;

%------------------------- (fin des donnees) ------------------------------

%--------------------------------------------------------------------------
% Determination des CL 
%--------------------------------------------------------------------------
if length(edgeCL)==size(edge,1)
    clD=1;  % CL de Dirichlet u=g sur tout le bord.
    fprintf('CL de Dirichlet (pression  moyenne nulle) / ');
else
    clD=0;  % CL mixtes Dirichlet/Neumann
    fprintf('CL mixtes Dirichlet/Neumann / ');
end

if sdr == 1
    fprintf('solveur direct\n');
else
    fprintf('solveur iteratif MINRES\n');
end

%--------------------------------------------------------------------------
% Solveur et CL.
%
% - Cas d'un solveur direct (sdr=1)
%   1. Pour des conditions limites de Dirichlet pures, la pression cherche 
%      est  moyenne nulle. Formulation avec multiplicateur de Lagrange 
%      pour la moyenne nulle de la pression (ml=1). Cette formulation 
%      conduit a une matrice modifie inversible.
%   2. Pour des conditions limites mixtes Dirichlet/Neumann, la pression 
%      cherche n'est pas  moyenne nulle -> pas de multiplicateur de 
%      Lagrange (ml=0). La matrice de Stokes est directement inversible.
%
% - Cas du solveur iteratif MINRES (sdr=0) :  pas de distinction de cas des
%   CL, pas de multiplicateur de Lagrange (ml=0).
%--------------------------------------------------------------------------
if (clD==1 && sdr==1)
    ml=1;  % multiplicateur de Lagrange pour la pression  moyenne nulle
    fprintf('->multiplicateur de Lagrange pour la pression  moyenne nulle \n');
else
    ml=0;  % pas de multiplicateur de Lagrange
end
fprintf('-------------------------------------------------------------\n');

%--------------------------------------------------------------------------
% Maillages P1 et P2
%--------------------------------------------------------------------------
outmesh=meshP1P2(node,edge,edgeCL,hdata,optionmesh);

v1=outmesh.v1; t1=outmesh.t1; e1=outmesh.e1;
isupp1=outmesh.isupp1; tsupp1=outmesh.tsupp1;
v2=outmesh.v2; t2=outmesh.t2; e2=outmesh.e2;

ib2=outmesh.ib2; ibT=outmesh.ibT; 

% nv1, nt1, ne1 : nb de noeuds, de triangles et d'aretes du maillage P1
% nv2, nt2, ne2 : nb de noeuds, de triangles et d'aretes du maillage P2
nv1=size(v1,1); nt1=size(t1,1); ne1=size(e1,1);
nv2=size(v2,1); nt2=size(t2,1); ne2=size(e2,1);

% Indices des degres de liberte (interieurs)  pour vitesse/pression et 
% eventuellement multiplicateur de Lagrange
icT=(1:(2*nv2+nv1+ml))'; 
icT(ibT)=[];      

fprintf('Maillage P1 : nombre de noeuds    (nv1) = %d\n',nv1);
fprintf('              nombre de triangles (nt1) = %d\n',nt1);
fprintf('Maillage P2 : nombre de noeuds    (nv2) = %d\n',nv2);
fprintf('              nombre de triangles (nt2) = %d\n',nt2);

% Affichage du maillage
patch('faces',t2(:,1:3),'vertices',v2(:,1:2),'facecolor','white','edgecolor','blue');
axis equal; axis off

%-------------------------------
% Calcul des matrices de Stokes   
%-------------------------------
tstart=tic;
fprintf('------------------------------------\n');
fprintf('Construction des matrices de Stokes \n');
fprintf('------------------------------------\n');
fprintf('Construction des matrices P2 : ');
[DxDx,DyDy,DxDy,DxW,DyW,SS]=matriceP2vect(v1,t1);
fprintf('Construction des matrices P1 : ');
[DxDxP1,DyDyP1,WW]=matriceP1vect(v1,t1);
fprintf('Temps d''assemblage des matrices P1 et P2: %f s\n',toc(tstart));

%Matrice de Stokes : formation D(u):D(v)
A=[(nu*2)*DxDx+nu*DyDy        nu*DxDy'             -DxW;
   nu*DxDy              nu*DxDx+(2*nu)*DyDy        -DyW;
   -DxW'                       -DyW'     sparse(nv1,nv1)];

% Pour le calcul de la moyenne de p (uniquement avec des CL de Dirichlet)
if clD==1
    w=zeros(nv1,1);
    for k=1:nt1
        pt=t1(k,1:3);
        aire=t1(k,5);
        w(pt)=w(pt)+aire/3;
    end
end
% CL de Dirichlet et solveur direct: multiplicateur de Lagrange pour la 
% pression (moyenne nulle)
if (clD==1 && sdr==1)
    C=[sparse(2*nv2,1);w/(w'*ones(nv1,1))];
    A=[A  C;
       C' 0];
end
%-----------------------------------
% Traitement des conditions limites
%-----------------------------------
tstart=tic;
% conditions limites u=g pour le bord 
xbd=v2(ib2,1);  % coordonnees des noeuds du bord 
ybd=v2(ib2,2);
[uxbd,uybd]=gcl(xbd,ybd);
ubd=[uxbd; uybd];

% Terme source
x=v2(:,1); y=v2(:,2); % coordonnees des noeuds 
[fx,fy]=fscmb(x,y);

% Second membre complet

b=-A(:,ibT)*ubd; 
ld=b+[SS*fx;SS*fy;sparse(nv1+ml,1)];
ld(ibT)=[];

% Mise  jour de la matrice A : on ne retient que les noeuds interieurs 
A=A(icT,icT);
fprintf('Taille de la matrice A : %d x %d (nnz = %d)\n',size(A), nnz(A));
fprintf('Temps de traitement des CL: %f s\n',toc(tstart));

%--------------------------------
% Résolution du systeme lineaire
%--------------------------------
tstart=tic;
if sdr==1
    % Solveur direct

    % sol=A\ld;

    % Remarque : si le systeme est tres grand, il faut mieux utiliser une
    % renumerotation par symamd (au lieu de amd par defaut).
    pt=symamd(A);          % re-ordering 'symetric approximate minimum degree'
    sol = A(pt,pt)\ld(pt); % resolution du systeme permute
    [~,invpt]=sort(pt);  
    sol=sol(invpt);        % equivalent a sol(pt)=sol, mais bcp plus rapide...
    fprintf('Solveur direct (avec rearrangement symamd)\n');
else
    % Solveur iteratif MINRES (avec preconditionnement)

    nbdlu = 2*nv2-length(ibT);
    diagA=diag(A(1:nbdlu,1:nbdlu));
    B=A(1:nbdlu,nbdlu+1:end);
    M1=diag([diagA; diag(B'*diag(sparse(1./diagA))*B)]); 
    M2=[];
    tol=1e-6;
    maxit=2000;
    [sol,flag,relres,iter] = minres(A,ld,tol,maxit,M1,M2);
    fprintf('Solveur MINRES : residu relatif : %g, nb d''iterations : %d\n',relres,iter);
end

% reconstruction de la vitesse et de la pression
u=zeros(2*nv2+nv1+ml,1);
u(icT)=sol;
u(ibT)=full(ubd);
Ux=u(1:nv2);
Uy=u((nv2+1):(2*nv2));
Pr=u((2*nv2+1):end-ml);    

% Calcul de la pression a moyenne nulle si solveur iteratif et CL Dirichlet
if (sdr==0 && clD==1)
    Pr=Pr-w'*Pr/(w'*ones(nv1,1));
end

fprintf('Temps de resolution du systeme lineaire: %f s\n',toc(tstart));

%-----------
% Affichage
%-----------
pmesh=true;  % affichage du maillage
plotting(Ux,Uy,Pr,t1,t2,v1,v2,pmesh);



